/// Provides interoperability with [Node.js](https://nodejs.org) APIs.
library nodejs_interop.node;

import 'dart:async';
import 'dart:js';
import 'js.dart';
import 'src/node/globals.dart';
import 'src/node/stream.dart';

export 'src/node/assert.dart';
export 'src/node/buffer.dart';
export 'src/node/child_process.dart';
export 'src/node/cluster.dart';
export 'src/node/console.dart';
export 'src/node/events.dart';
export 'src/node/fs.dart';
export 'src/node/globals.dart';
export 'src/node/os.dart';
export 'src/node/path.dart';
export 'src/node/stream.dart';
export 'src/node/util.dart';
export 'src/node/vm.dart';

/// Converts the specified [readable] stream to a [Stream].
Stream<T> convertToStream<T>(Readable<T> readable) {
  var controller = new StreamController<T>();
  readable
    ..on('error', allowInterop((error) => controller..addError(error)..close()))
    ..on('data', allowInterop((data) => controller.add(data)))
    ..on('end', allowInterop(() => controller.close()));

  return controller.stream;
}

/// Requires the JavaScript module with the specified [identifier].
///
/// Throws a [DeferredLoadException] if the module could not be required.
/// If a [onError] handler is provided, it is called with the [identifier] as argument, and its return value is used instead.
///
/// This function is prefered to the [require] function because it properly exposes to the Dart engine the types exported by the required module.
T loadLibrary<T>(String identifier, {T onError(String identifier)}) {
  try {
    var module = '\$${identifier.replaceAll(new RegExp('[-/@.]'), '_')}';
    if (context[module] == null) context[module] = (context['require'] as JsFunction).apply([identifier]);
    return require(identifier);
  }

  on JSError {
    if (onError != null) return onError(identifier);
    throw new DeferredLoadException(identifier);
  }
}
