/// Provides interoperability with [`assert`](https://nodejs.org/api/assert.html) module.
@JS()
library nodejs_interop.node.assertion;

import 'dart:js';
import 'package:js/js.dart';
import 'package:nodejs_interop/js.dart';

/// The [`assert`](https://nodejs.org/api/assert.html) module.
@JS()
abstract class AssertModule implements Function {

  /// An alias for the [ok] method.
  external void call(value, [String message]);

  /// Tests for deep equality between the [actual] and [expected] parameters.
  external void deepEqual(actual, expected, [String message]);

  /// Tests for deep strict equality between the [actual] and [expected] parameters.
  external void deepStrictEqual(actual, expected, [String message]);

  /// Asserts that the function [block] does not throw an [error].
  external void doesNotThrow(Function block, [error, String message]);

  /// Tests shallow, coercive equality between the [actual] and [expected] parameters using the abstract equality comparison (`==`).
  external void equal(actual, expected, [String message]);

  /// Throws an assertion error.
  external void fail([actualOrMessage, expected, String message, String operator, JsFunction stackStartFunction]);

  /// Throws [value] if [value] is truthy. This is useful when testing the `error` argument in callbacks.
  external void ifError(value);

  /// Tests for any deep inequality. Opposite of [deepEqual].
  external void notDeepEqual(actual, expected, [String message]);

  /// Tests for deep strict inequality. Opposite of [deepStrictEqual].
  external void notDeepStrictEqual(actual, expected, [String message]);

  /// Tests shallow, coercive inequality with the abstract equality comparison (`!=`).
  external void notEqual(actual, expected, [String message]);

  /// Tests strict inequality as determined by the strict equality comparison (`!==`).
  external void notStrictEqual(actual, expected, [String message]);

  /// Tests if [value] is truthy.
  external void ok(value, [String message]);

  /// Tests strict equality as determined by the strict equality comparison (`===`).
  external void strictEqual(actual, expected, [String message]);

  /// Expects the function [block] to throw an [error].
  external void throws(Function block, [error, String message]);
}

/// The context of an assertion error.
@JS()
@anonymous
class AssertionErrorOptions {

  /// Creates a new options object.
  external factory AssertionErrorOptions({actual, expected, String message, String operator, JsFunction stackStartFunction});

  /// The actual value.
  external dynamic get actual;

  /// The expected value.
  external dynamic get expected;

  /// The error message.
  external String get message;

  /// The comparison operator used by the assertion.
  external String get operator;

  /// The constructor function used to customize the stack trace.
  external JsFunction get stackStartFunction;
}

/// Indicates the failure of an assertion.
/// Such errors commonly indicate inequality of actual and expected value.
@JS(r'$assert.AssertionError')
class JSAssertionError extends JSError {

  /// Creates a new assertion error.
  external JSAssertionError(AssertionErrorOptions options);

  /// The actual value.
  external dynamic get actual;

  /// The expected value.
  external dynamic get expected;

  /// Value indicating whether the error [message] is generated.
  external bool get generatedMessage;

  /// The comparison operator used by the assertion.
  external String get operator;
}
