/// Provides interoperability with [`child_process`](https://nodejs.org/api/child_process.html) module.
@JS()
library nodejs_interop.node.child_process;

import 'package:js/js.dart';
import 'package:nodejs_interop/js.dart';
import 'package:nodejs_interop/src/node/events.dart';
import 'package:nodejs_interop/src/node/stream.dart';

/// Represents a spawned child process.
@JS()
abstract class ChildProcess extends EventEmitter {

  /// A reference to this child's IPC channel. If no IPC channel currently exists, this property is `null`.
  external dynamic get channel;

  /// Value indicating whether it is still possible to send and receive messages from this child process.
  /// When [connected] is `false`, it is no longer possible to send or receive messages.
  external bool get connected;

  /// Value indicating whether this child process successfully received a signal from [kill].
  /// The [killed] property does not indicate that the child process has been terminated.
  external bool get killed;

  /// The process identifier (PID) of this child process.
  external int get pid;

  /// A [Readable] stream that represents this child process's `stderr`.
  /// [stderr] is an alias for `stdio[2]`. Both properties will refer to the same value.
  external Readable get stderr;

  /// A [Writable] stream that represents this child process's `stdin`.
  /// [stdin] is an alias for `stdio[0]`. Both properties will refer to the same value.
  external Writable get stdin;

  /// A sparse array of pipes to this child process, corresponding with positions in the `stdio` option
  /// passed to [ChildProcessModule.spawn] that have been set to the value `"pipe"`.
  ///
  /// Note that `stdio[0]`, `stdio[1]`, and `stdio[2]` are also available as [stdin], [stdout], and [stderr], respectively.
  external List get stdio;

  /// A [Readable] stream that represents this child process's `stdout`.
  /// [stdout] is an alias for `stdio[1]`. Both properties will refer to the same value.
  external Readable get stdout;

  /// Closes the IPC channel between parent and child, allowing the child to exit gracefully once there are no other connections keeping it alive.
  ///
  /// After calling this method the [connected] properties in both the parent and child (respectively) will be set to `false`,
  /// and it will be no longer possible to pass messages between the processes.
  external void disconnect();

  /// Sends a [signal] to this child process.
  /// Note that while the method is called `kill`, the signal delivered to the child process may not actually terminate the process.
  external void kill([String signal = 'SIGTERM']);

  /// When an IPC channel has been established between the parent and child (i.e. when using [ChildProcessModule.fork]),
  /// the [send] method can be used to send messages to the child process. When the child process is a Node.js instance,
  /// these messages can be received via the `process.on("message")` event.
  ///
  /// The message goes through JSON serialization and parsing. The resulting message might not be the same as what is originally sent.
  external bool send(message, [sendHandleOrOptions, optionsOrCallback, CompletionCallback0 callback]);
}

/// The [`child_process`](https://nodejs.org/api/child_process.html) module.
@JS()
abstract class ChildProcessModule {

  /// Spawns a shell then executes the [command] within that shell, buffering any generated output.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  ///
  /// If a [callback] function is provided, it is called with the arguments `(error, stdout, stderr)`.
  /// The `stdout` and `stderr` arguments passed to the [callback] will contain the `stdout` and `stderr` output of the child process.
  external ChildProcess exec(String command, [optionsOrCallback, CompletionCallback2 callback]);

  /// Spawns the specified [executable] file, with command line arguments in `args`.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  ///
  /// If a [callback] function is provided, it is called with the arguments `(error, stdout, stderr)`.
  /// The `stdout` and `stderr` arguments passed to the [callback] will contain the `stdout` and `stderr` output of the child process.
  external ChildProcess execFile(String executable, [argsOrOptions, optionsOrCallback, CompletionCallback2 callback]);

  /// Spawns the specified [executable] file, with command line arguments in `args`. Returns the `stdout` from the command.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  ///
  /// This is a synchronous call and will block until the child process terminates.
  /// If the process times out, or has a non-zero exit code, this method will throw.
  external dynamic execFileSync(String executable, [argsOrOptions, SpawnOptions options]);

  /// Spawns a shell then executes the [command] within that shell, buffering any generated output. Returns the `stdout` from the command.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  ///
  /// This is a synchronous call and will block until the child process terminates.
  /// If the process times out, or has a non-zero exit code, this method will throw.
  external dynamic execSync(String command, [SpawnOptions options]);

  /// Spawns a Node.js process using the given [modulePath], with command line arguments in `args`.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  ///
  /// Returns a [ChildProcess] that has an additional IPC communication channel built-in allowing messages
  /// to be passed back and forth between the parent and child.
  external ChildProcess fork(String modulePath, [argsOrOptions, SpawnOptions options]);

  /// Spawns a new process using the given [executable], with command line arguments in `args`.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  external ChildProcess spawn(String executable, [argsOrOptions, SpawnOptions options]);

  /// Spawns a new process using the given [executable], with command line arguments in `args`.
  /// [SpawnOptions] may be passed to customize how the process is spawned.
  ///
  /// This is a synchronous call and will block until the child process terminates.
  external SpawnResult spawnSync(String executable, [argsOrOptions, SpawnOptions options]);
}

/// Options used to customize the sending of a message via an IPC channel.
@JS()
@anonymous
class SendOptions {

  /// Creates a new options object.
  external factory SendOptions({bool keepOpen = false});

  /// Value indicating whether the socket is kept open in the sending process.
  external bool get keepOpen;
}

/// Options used to customize the spawning of a child process.
@JS()
@anonymous
class SpawnOptions {

  /// Creates a new options object.
  external factory SpawnOptions({
    String argv0,
    String cwd,
    bool detached = false,
    String encoding,
    env,
    List<String> execArgv,
    String execPath,
    int gid,
    input,
    killSignal = 'SIGTERM',
    int maxBuffer = 200 * 1024,
    shell,
    bool silent = false,
    stdio = 'pipe',
    int timeout = 0,
    int uid,
    bool windowsHide = false,
    bool windowsVerbatimArguments = false
  });

  /// The value of `argv[0]` sent to the child process.
  /// Only supported by the [ChildProcessModule.spawn] method.
  external String get argv0;

  /// The working directory for the child process.
  external String get cwd;

  /// Value indicating whether to prepare child to run independently of its parent process.
  /// Only supported by the [ChildProcessModule.spawn] method.
  external bool get detached;

  /// The encoding used for decoding `stdout` and `stderr`.
  /// Not supported by the [ChildProcessModule.fork] and [ChildProcessModule.spawn] methods.
  external String get encoding;

  /// The environment variables for the child process.
  external dynamic get env;

  /// The list of arguments passed to the executable.
  /// Only supported by the [ChildProcessModule.fork] method.
  external List<String> get execArgv;

  /// The executable used to create the child process.
  /// Only supported by the [ChildProcessModule.fork] method.
  external String get execPath;

  /// The group identity of the child process.
  external int get gid;

  /// The value which will be passed as `stdin` to the spawned process. Supplying this value will override `stdio[0]`.
  /// Only supported by the [ChildProcessModule]'s synchronous methods.
  external dynamic get input;

  /// The signal value to be used when the spawned process will be killed.
  /// Not supported by the [ChildProcessModule.fork] and [ChildProcessModule.spawn] methods.
  external dynamic get killSignal;

  /// The largest amount of data in bytes allowed on `stdout` or `stderr`.
  /// Not supported by the [ChildProcessModule.fork] and [ChildProcessModule.spawn] methods.
  external int get maxBuffer;

  /// Value indicating whether to run the command inside a shell. A different shell can be specified as a string.
  /// Not supported by the [ChildProcessModule.execFile], [ChildProcessModule.execFileSync] and [ChildProcessModule.fork] methods.
  external dynamic get shell;

  /// Value indicating whether `stdio` of the child process will be piped to the parent instead of be inherited from.
  /// Only supported by the [ChildProcessModule.fork] method.
  external bool get silent;

  /// The `stdio` configuration of the child process.
  /// Not supported by the [ChildProcessModule.exec] and [ChildProcessModule.execFile] methods.
  external dynamic get stdio;

  /// The maximum amount of time, in milliseconds, the child process is allowed to run.
  /// Not supported by the [ChildProcessModule.fork] and [ChildProcessModule.spawn] methods.
  external int get timeout;

  /// The user identity of the child process.
  external int get uid;

  /// Value indicating whether to hide the subprocess console window that would normally be created on Windows systems.
  /// Not supported by the [ChildProcessModule.fork] method.
  external bool get windowsHide;

  /// Value indicating whether no quoting or escaping of arguments is done on Windows.
  /// Not supported by the [ChildProcessModule.exec], [ChildProcessModule.execFileSync] and [ChildProcessModule.execSync] methods.
  external bool get windowsVerbatimArguments;
}

/// Represents the result of spawning a child process.
@JS()
abstract class SpawnResult {

  /// The error object if the child process failed or timed out.
  external JSError get error;

  /// The list of results from `stdio` output.
  external List get output;

  /// The process identifier (PID) of the child process.
  external int get pid;

  /// The signal used to kill the child process.
  external String get signal;

  /// The exit code of the child process.
  external int get status;

  /// The contents of `output[1]`.
  external dynamic get stderr;

  /// The contents of `output[2]`.
  external dynamic get stdout;
}
