/// Provides interoperability with [`cluster`](https://nodejs.org/api/cluster.html) module.
@JS()
library nodejs_interop.node.cluster;

import 'package:js/js.dart';
import 'package:nodejs_interop/js.dart';
import 'package:nodejs_interop/src/node/child_process.dart';
import 'package:nodejs_interop/src/node/events.dart';

/// The [`cluster`](https://nodejs.org/api/cluster.html) module.
@JS()
abstract class ClusterModule extends EventEmitter {

  /// No scheduling policy: left to the operating system.
  external static int get SCHED_NONE;

  /// The round-robin scheduling policy.
  external static int get SCHED_RR;

  /// Value indicating whether the process is a master.
  external bool get isMaster;

  /// Value indicating whether the process is a worker (i.e. not a master).
  external bool get isWorker;

  /// The scheduling policy, either [SCHED_RR] for round-robin or [SCHED_NONE] to leave it to the operating system.
  /// This is a global setting and effectively frozen once either the first worker is spawned, or [setupMaster] is called, whichever comes first.
  external int get schedulingPolicy;
  external set schedulingPolicy(int value);

  /// An object containing the cluster settings, including the default values.
  external ClusterSettings get settings;

  /// A reference to the current worker object.
  /// This property is not available in the master process.
  external Worker get worker;

  /// An object that stores the active worker objects, keyed by [Worker.id] field.
  /// This property can only be called from the master process.
  external dynamic get workers;

  /// Calls [Worker.disconnect] on each worker in [workers].
  ///
  /// This method takes an optional [callback] argument which will be called when finished.
  /// This method can only be called from the master process.
  external void disconnect([CompletionCallback0 callback]);

  /// Spawns a new worker process, optionally adding [environment] to the worker process environment.
  /// This method can only be called from the master process.
  external Worker fork([environment]);

  /// Changes the default [fork] behavior. Once called, the settings will be present in [settings].
  /// This method can only be called from the master process.
  external void setupMaster([ClusterSettings settings]);
}

/// Provides the settings of a cluster.
@JS()
@anonymous
class ClusterSettings {

  /// Creates new cluster settings.
  external factory ClusterSettings({
    List<String> args,
    String exec,
    List<String> execArgv,
    int gid,
    inspectPort,
    List stdio,
    bool silent = false,
    int uid
  });

  /// String arguments passed to worker. Defaults to `process.argv.slice(2)`.
  external List<String> get args;

  /// File path to worker file. Defaults to `process.argv[1]`.
  external String get exec;

  /// List of string arguments passed to the Node.js executable. Defaults to `process.execArgv`.
  external List<String> get execArgv;

  /// The group identity of the process.
  external int get gid;

  /// The inspector port of worker. This can be an integer, or a function that takes no arguments and returns an integer.
  external dynamic get inspectPort;

  /// Value indicating whether to send output to parent's `stdio`.
  external bool get silent;

  /// The `stdio` configuration of the forked processes.
  external List get stdio;

  /// The user identity of the process.
  external int get uid;
}

/// Contains all public information and methods about a worker.
///
/// In the master it can be obtained using [ClusterModule.workers].
/// In a worker it can be obtained using [ClusterModule.worker].
@JS()
abstract class Worker extends EventEmitter {

  /// Value allowing to distinguish between voluntary and accidental exit.
  /// The master may choose not to respawn a worker based on this value.
  external bool get exitedAfterDisconnect;

  /// The worker identifier.
  external int get id;

  /// The underlying child process.
  external ChildProcess get process;

  /// In a worker, this method will close all servers, wait for the `close` event on those servers, and then disconnect the IPC channel.
  /// In the master, an internal message is sent to the worker causing it to call [disconnect] on itself.
  external Worker disconnect();

  /// Returns a value indicating whether the worker is connected to its master via its IPC channel.
  /// A worker is connected to its master after it has been created. It is disconnected after the `disconnect` event is emitted.
  external bool isConnected();

  /// Returns a value indicating whether the worker's process has terminated (either because of exiting or being signaled).
  external bool isDead();

  /// Kills the worker.
  ///
  /// In the master, it does this by disconnecting the [process], and once disconnected, killing with [signal].
  /// In the worker, it does it by disconnecting the channel, and then exiting with code `0`.
  external void kill([String signal = 'SIGTERM']);

  /// Sends a message to a worker or master, optionally with a handle.
  ///
  /// In the master this sends a message to a specific worker. It is identical to [ChildProcess.send].
  /// In a worker this sends a message to the master. It is identical to `process.send()`.
  external bool send(message, [sendHandleOrCallback, CompletionCallback0 callback]);
}
