/// Provides interoperability with [`util`](https://nodejs.org/api/util.html) module.
@JS()
library nodejs_interop.node.util;

import 'package:js/js.dart';
import 'package:nodejs_interop/js.dart';

/// Interface defining a logging function.
typedef void LoggingFunction(data, [arg1, arg2, arg3, arg4, arg5, arg6, arg7]);

/// The [`util`](https://nodejs.org/api/util.html) module.
@JS()
abstract class UtilModule {

  /// Takes an `async` [function] (or a function that returns a [Promise])
  /// and returns a function following the Node.js error first callback style.
  ///
  /// In the callback, the first argument will be the rejection reason (or `null` if the promise resolved),
  /// and the second argument will be the resolved value.
  external Function callbackify(Function function);

  /// Creates a function that conditionally writes debug messages to `stderr` based on the existence of the `NODE_DEBUG` environment variable.
  /// If the [section] name appears within the value of that environment variable, then the returned function operates similar to `Console.error()`.
  /// If not, then the returned function is a no-op.
  external LoggingFunction debuglog(String section);

  /// Wraps the given [function] in such a way that it is marked as deprecated.
  /// Returns a function that will emit a `DeprecationWarning` using the `process.on("warning")` event.
  ///
  /// By default, the deprecation warning will be emitted and printed to `stderr` exactly once, the first time the returned function is called.
  /// After this warning is emitted, the wrapped function is called.
  external Function deprecate(Function function, [String message]);

  /// Returns a formatted string using the first argument as a `printf`-like format.
  ///
  /// The first argument is a string containing zero or more placeholder tokens.
  /// Each placeholder token is replaced with the converted value from the corresponding argument. Supported placeholders are:
  /// - `%s`: String.
  /// - `%d`: Number.
  /// - `%i`: Integer.
  /// - `%f`: Floating point value.
  /// - `%j`: JSON. Replaced with the string `"[Circular]"` if the argument contains circular references.
  /// - `%o`: Object. A string representation of an object with generic formatting. This will show the full object including non-enumerable symbols and properties.
  /// - `%O`: Object. A string representation of an object with generic formatting. This will show the full object excluding non-enumerable symbols and properties.
  /// - `%%`: Single percent sign (`"%"`). This does not consume an argument.
  external String format(String format, [arg1, arg2, arg3, arg4, arg5, arg6, arg7]);

  /// Returns a string representation of [object] that is primarily useful for debugging.
  /// Additional [options] may be passed that alter certain aspects of the formatted string.
  external String inspect(object, [InspectOptions options]); // TODO: an instance of InspectFunction, a callable class with properties.

  /// Takes a [function] following the common Node.js callback style, i.e. taking a `(err, value) => ...` callback as the last argument,
  /// and returns a version that returns [Promise]s.
  external Function promisify(Function function);
}

/// Options used to alter certain aspects of a formatted string.
@JS()
@anonymous
class InspectOptions {

  /// Creates a new options object.
  external factory InspectOptions({
    int breakLength = 60,
    bool colors = false,
    bool customInspect = true,
    int depth = 2,
    int maxArrayLength = 100,
    bool showHidden = false,
    bool showProxy = false
  });

  /// The length at which an object's keys are split across multiple lines.
  /// Set to [double.INFINITY] to format an object as a single line.
  external int get breakLength;

  /// Value indicating whether the output will be styled with ANSI color codes.
  external bool get colors;

  /// Value indicating whether custom inspection functions exported on the object being inspected will be called.
  external bool get customInspect;

  /// Specifies the number of times to recurse while formatting the object.
  /// To make it recurse indefinitely pass `null`.
  external int get depth;

  /// Specifies the maximum number of list elements to include when formatting.
  /// Set to `null` to show all list elements. Set to `0` or negative to show no list elements.
  external int get maxArrayLength;

  /// Value indicating whether the object's non-enumerable symbols and properties will be included in the formatted result.
  external bool get showHidden;

  /// Value indicating whether objects and functions that are proxy objects will be introspected to show their target and handler objects.
  external bool get showProxy;
}
