/// Provides interoperability with [`vm`](https://nodejs.org/api/vm.html) module.
@JS()
library nodejs_interop.node.vm;

import 'package:js/js.dart';
import 'package:nodejs_interop/src/node/buffer.dart';

/// Represents a precompiled script that can be executed in specific sandboxes (or "contexts").
@JS(r'$vm.Script')
class Script {

  /// Creates a new script from the specified [code].
  external factory Script(String code, [ScriptOptions options]);

  /// Runs the compiled code contained by this script within the context of the [contextifiedSandbox] and returns the result.
  /// Running code does not have access to the local scope.
  external dynamic runInContext(contextifiedSandbox, [ContextOptions options]);

  /// Contextifies the given `sandbox`, runs the compiled code contained by this script within the created sandbox, and returns the result.
  /// Running code does not have access to the local scope.
  external dynamic runInNewContext([sandboxOrOptions, ContextOptions options]);

  /// Runs the compiled code contained by this script within the context of the current `global` object and returns the result.
  /// Running code does not have access to local scope, but does have access to the current `global` object.
  external dynamic runInThisContext([ContextOptions options]);
}

/// Options used to customize the compilation of JavaScript code.
@JS()
@anonymous
class ContextOptions {

  /// Creates a new options object.
  external factory ContextOptions({
    bool breakOnSigint,
    int columnOffset,
    bool displayErrors,
    String filename,
    int lineOffset,
    int timeout
  });

  /// Value indicating whether the execution will be terminated when `SIGINT` (i.e. Ctrl+C) is received.
  /// Only supported by the [Script.runInContext] method.
  external bool get breakOnSigint;

  /// The column number offset that is displayed in stack traces produced by the script.
  external int get columnOffset;

  /// Value indicating whether when an error occurs while compiling the code, the line of code causing the error is attached to the stack trace.
  external bool get displayErrors;

  /// The filename used in stack traces produced by the script.
  external String get filename;

  /// The line number offset that is displayed in stack traces produced by the script.
  external int get lineOffset;

  /// The number of milliseconds to execute code before terminating execution.
  external int get timeout;
}

/// Options used to customize the compilation of JavaScript code.
@JS()
@anonymous
class ScriptOptions extends ContextOptions {

  /// Creates a new options object.
  external factory ScriptOptions({
    Buffer cachedData,
    int columnOffset,
    bool displayErrors,
    String filename,
    int lineOffset,
    bool produceCachedData,
    int timeout
  });

  /// Provides an optional [Buffer] with V8's code cache data for the supplied source.
  external Buffer get cachedData;

  /// Value indicating whether V8 will attempt to produce code cache data if no [cachedData] is present.
  external bool get produceCachedData;
}

/// The [`vm`](https://nodejs.org/api/vm.html) module.
@JS()
abstract class VMModule {

  /// If given a [sandbox] object, the [createContext] method will prepare that sandbox so that it can be used
  /// in calls to [runInContext] or [Script.runInContext]. Inside such scripts, the [sandbox] object will be the global object,
  /// retaining all of its existing properties but also having the built-in objects and functions any standard global object has.
  /// Outside of scripts run by the `vm` module, global variables will remain unchanged.
  ///
  /// If [sandbox] is omitted, a new empty contextified sandbox object will be returned.
  external dynamic createContext([sandbox]);

  /// Returns a value indicating whether the given [sandbox] object has been contextified using [createContext].
  external bool isContext(sandbox);

  /// Compiles [code], runs it within the context of the [contextifiedSandbox], then returns the result.
  /// Running code does not have access to the local scope.
  external dynamic runInContext(String code, contextifiedSandbox, [ContextOptions options]);

  /// Contextifies the given `sandbox` (or creates a new one), compiles the [code], runs it within the created context, then returns the result.
  /// Running code does not have access to the local scope.
  external dynamic runInNewContext(String code, [sandboxOrOptions, ContextOptions options]);

  /// Compiles [code], runs it within the context of the current `global` and returns the result.
  /// Running code does not have access to local scope, but does have access to the current `global` object.
  external dynamic runInThisContext(String code, [ContextOptions options]);
}
