import 'dart:js';
import 'package:nodejs_interop/io.dart';
import 'package:test/test.dart';

/// Returns a list of a given [object]'s own enumerable properties.
List<String> getKeysOfObject(JsObject object) => (context['Object'] as JsFunction).callMethod('keys', [object]);

/// Tests the features of the [Platform] class.
void main() => group('Platform', () {
  final isMacOS = context['process']['platform'] == 'darwin';
  final isWindows = context['process']['platform'] == 'win32';

  group('.environment', () {
    var env = const Platform().environment;
    JsObject jsEnv = context['process']['env'];

    test('should reflect the values of the `process.env` property', () {
      var keys = getKeysOfObject(jsEnv).where((key) => !key.contains('__dart__'));
      expect(env.length, equals(keys.length));
      for (var key in keys) expect(env[key], equals(jsEnv[key]), reason: key);
    });

    test('should be an unmodifiable map', () {
      expect(() => env['Foo'] = 'Bar', throwsUnsupportedError);
    });

    test('should be case-insensitive on Windows, case-sensitive on other platforms', () {
      expect(env['home'], isWindows ? isNotEmpty : isNull);
    });
  });

  group('.executable', () {
    test('should be `node.exe` on Windows and `node` on other platforms', () {
      expect(const Platform().executable, endsWith(isWindows ? 'node.exe' : 'node'));
    });
  });

  group('.localeName', () {
    test('should return a locale formatted like "<language>_<REGION>"', () {
      expect(const Platform().localeName, matches(new RegExp(r'^[a-z]{2,3}_[A-Z]{2,3}$')));
    });
  });

  group('.operatingSystem', () {
    test('should be `macos` on macOS, `windows` on Windows and `linux` on other platforms', () {
      expect(const Platform().operatingSystem, endsWith(isWindows ? 'windows' : (isMacOS ? 'macos' : 'linux')));
    });
  });

  group('.pathSeparator', () {
    test('should be `\\` on Windows and `/` on other platforms', () {
      expect(const Platform().pathSeparator, equals(isWindows ? r'\' : '/'));
    });
  });

  group('.resolvedExecutable', () {
    test('should be the absolute path to `node.exe` on Windows and `node` on other platforms', () {
      expect(const Platform().resolvedExecutable, endsWith(isWindows ? r'\node.exe' : '/node'));
    });
  });

  group('.version', () {
    test('should be a SemVer string representing the `process.version` property', () {
      expect(const Platform().version, equals(context['process']['version'].substring(1)));
    });
  });
});
