import 'dart:js';
import 'package:nodejs_interop/nodejs_interop.dart';
import 'package:test/test.dart';

/// Tests the features of the [AssertModule] class.
void main() {
  final _assert = loadLibrary<AssertModule>('assert');

  group('AssertModule', () {
    group('.deepEqual()', () {
      test('should not throw an assertion error if the actual and expected value are deeply equal', () {
        var map = new JsObject.jsify({'foo': 123});
        expect(() => _assert.deepEqual(map, new JsObject.jsify({'foo': '123'})), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are not deeply equal', () {
        var map = new JsObject.jsify({'foo': 'bar'});
        expect(() => _assert.deepEqual(map, new JsObject.jsify({'foo': 'qux'})), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.deepStrictEqual()', () {
      test('should not throw an assertion error if the actual and expected value are deeply and strictly equal', () {
        var map = new JsObject.jsify({'foo': '123'});
        expect(() => _assert.deepStrictEqual(map, new JsObject.jsify({'foo': '123'})), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are not deeply and strictly equal', () {
        var map = new JsObject.jsify({'foo': '123'});
        expect(() => _assert.deepStrictEqual(map, new JsObject.jsify({'foo': 123})), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.doesNotThrow()', () {
      test('should not throw an assertion error if the block does not throw an error', () {
        var block = allowInterop(() => null);
        expect(() => _assert.doesNotThrow(block), returnsNormally);
      });

      test('should throw an assertion error if the block throws an error', () {
        var block = allowInterop(() => throw new Error());
        expect(() => _assert.doesNotThrow(block), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.equal()', () {
      test('should not throw an assertion error if the actual and expected value are equal', () {
        expect(() => _assert.equal(123, '123'), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are not equal', () {
        expect(() => _assert.equal('foo', 'bar'), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.fail()', () {
      test('should throw an assertion error', () {
        expect(_assert.fail, throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.ifError()', () {
      test('should not throw an assertion error if the value is falsy', () {
        expect(() => _assert.ifError(null), returnsNormally);
      });

      test('should throw an assertion error if the value is truthy', () {
        expect(() => _assert.ifError(new JSError()), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.notDeepEqual()', () {
      test('should not throw an assertion error if the actual and expected value are not deeply equal', () {
        var map = new JsObject.jsify({'foo': 'bar'});
        expect(() => _assert.notDeepEqual(map, new JsObject.jsify({'foo': 'qux'})), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are deeply equal', () {
        var map = new JsObject.jsify({'foo': 123});
        expect(() => _assert.notDeepEqual(map, new JsObject.jsify({'foo': '123'})), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.notDeepStrictEqual()', () {
      test('should not throw an assertion error if the actual and expected value are not deeply and strictly equal', () {
        var map = new JsObject.jsify({'foo': '123'});
        expect(() => _assert.notDeepStrictEqual(map, new JsObject.jsify({'foo': 123})), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are deeply and strictly equal', () {
        var map = new JsObject.jsify({'foo': '123'});
        expect(() => _assert.notDeepStrictEqual(map, new JsObject.jsify({'foo': '123'})), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.notEqual()', () {
      test('should not throw an assertion error if the actual and expected value are not equal', () {
        expect(() => _assert.notEqual('foo', 'bar'), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are equal', () {
        expect(() => _assert.notEqual(123, '123'), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.notStrictEqual()', () {
      test('should not throw an assertion error if the actual and expected value are not strictly equal', () {
        expect(() => _assert.notStrictEqual(123, '123'), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are strictly equal', () {
        expect(() => _assert.notStrictEqual('123', '123'), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.ok()', () {
      test('should not throw an assertion error if the value is truthy', () {
        expect(() => _assert(true), returnsNormally);
        expect(() => _assert.ok(true), returnsNormally);
      });

      test('should throw an assertion error if the value is falsy', () {
        expect(() => _assert(false), throwsA(const isInstanceOf<JSAssertionError>()));
        expect(() => _assert.ok(false), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.strictEqual()', () {
      test('should not throw an assertion error if the actual and expected value are strictly equal', () {
        expect(() => _assert.strictEqual('123', '123'), returnsNormally);
      });

      test('should throw an assertion error if the actual and expected value are not strictly equal', () {
        expect(() => _assert.strictEqual(123, '123'), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });

    group('.throws()', () {
      test('should not throw an assertion error if the block throws an error', () {
        var block = allowInterop(() => throw new Error());
        expect(() => _assert.ok(block), returnsNormally);
      });

      test('should throw an assertion error if the block does not throw an error', () {
        var block = allowInterop(() => null);
        expect(() => _assert.throws(block), throwsA(const isInstanceOf<JSAssertionError>()));
      });
    });
  });

  group('JSAssertionError', () {
    test('should be an instance of the `Exception` class', () {
      var error = new JSAssertionError(new AssertionErrorOptions(actual: 123, message: 'AssertionError'));
      expect(error, const isInstanceOf<Exception>());
      expect(error.actual, equals(123));
      expect(error.message, equals('AssertionError'));
    });
  });
}
